# from django.contrib.admin.utils import quote
# from django.contrib.auth.decorators import login_required
# from django.core.exceptions import PermissionDenied
# from django.http.response import HttpResponseRedirect
# from django.utils.decorators import method_decorator
# from django.utils.functional import cached_property
from django.utils.text import capfirst

# from django.utils.translation import gettext_lazy as _
from django.views.generic.base import ContextMixin, TemplateResponseMixin, View

# from django.views.generic.detail import BaseDetailView, SingleObjectMixin
# from django.views.generic.edit import BaseCreateView, BaseDeleteView, BaseUpdateView
# from django.views.generic.list import MultipleObjectMixin

# from django_filters.views import FilterMixin

# from . import consts, messages, utils


class SiteContext(ContextMixin):
    title = ""
    subtitle = ""
    meta_title = None
    meta_description = None

    def get_title(self):
        return capfirst(self.title)

    def get_subtitle(self):
        return capfirst(self.subtitle)

    def get_meta_title(self):
        return self.meta_title or self.get_page_title()

    def get_meta_description(self):
        return self.meta_description or self.get_page_title()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context = {
            "view": self,
            "title": self.get_title(),
            "subtitle": self.get_subtitle(),
            "meta_title": self.get_meta_title(),
            "meta_decription": self.get_meta_description(),
        }
        context.update(kwargs)
        return context


class SiteView(ContextMixin, TemplateResponseMixin, View):
    def has_permission(self, request):
        pass

    def dispatch(self, request, *args, **kwargs):
        self.has_permission(request)
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)



from django.contrib.auth import get_permission_codename
from django.contrib.auth.models import Permission


class PermissionHelper:
    """
    Provides permission-related helper functions to help determine what a
    user can do with a 'typical' model (where permissions are granted
    model-wide), and to a specific instance of that model.
    """

    def __init__(self, model):
        self.model = model
        self.opts = model._meta

    def get_all_model_permissions(self):
        """
        Return a queryset of all Permission objects pertaining to the `model`
        specified at initialisation.
        """

        return Permission.objects.filter(
            content_type__app_label=self.opts.app_label,
            content_type__model=self.opts.model_name,
        )

    def get_perm_codename(self, action):
        return get_permission_codename(action, self.opts)

    def user_has_specific_permission(self, user, perm_codename):
        """
        Combine `perm_codename` with `self.opts.app_label` to call the provided
        Django user's built-in `has_perm` method.
        """
        return user.has_perm("%s.%s" % (self.opts.app_label, perm_codename))

    def user_has_any_permissions(self, user):
        """
        Return a boolean to indicate whether `user` has any model-wide
        permissions
        """
        for perm in self.get_all_model_permissions().values("codename"):
            if self.user_has_specific_permission(user, perm["codename"]):
                return True
        return False

    def user_can_list(self, user):
        """
        Return a boolean to indicate whether `user` is permitted to access the
        list view for self.model
        """
        return self.user_has_any_permissions(user)

    def user_can_create(self, user):
        """
        Return a boolean to indicate whether `user` is permitted to create new
        instances of `self.model`
        """
        perm_codename = self.get_perm_codename("add")
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_inspect_obj(self, user, obj):
        """
        Return a boolean to indicate whether `user` is permitted to 'inspect'
        a specific `self.model` instance.
        """
        return self.user_has_any_permissions(user)

    def user_can_edit_obj(self, user, obj):
        """
        Return a boolean to indicate whether `user` is permitted to 'change'
        a specific `self.model` instance.
        """
        perm_codename = self.get_perm_codename("change")
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_delete_obj(self, user, obj):
        """
        Return a boolean to indicate whether `user` is permitted to 'delete'
        a specific `self.model` instance.
        """
        perm_codename = self.get_perm_codename("delete")
        return self.user_has_specific_permission(user, perm_codename)

    def user_can_unpublish_obj(self, user, obj):
        return False

    def user_can_copy_obj(self, user, obj):
        return False

    def user_is_owner_or_admin(self, user, obj, owner_field):
        obj_owner = getattr(obj, owner_field, None)
        if (obj_owner and obj_owner is user) or user.is_superadmin:
            return True
        else:
            return False

    def user_is_member(self, group_name):
        """Return a boolean to indicate whether `user` is a member of a
        specific `group` instance."""
        return False



"""
Routers provide a convenient and consistent way of automatically
determining the URL conf for your Web.

They are used by simply instantiating a Router class, and then registering
all the required ViewSets with that router.

For example, you might have a `urls.py` that looks something like this:

    router = DefaultRouter()
    router.register(UserViewSet)
    router.register(AccountViewsetGroup)
    urlpatterns = router.urls

"""
from inspect import isclass

from django.core.exceptions import ImproperlyConfigured
from django.urls.conf import include, path
from django.views import View

from django_hookup import core as hookup

from .settings import routers_settings
from .views import SiteView


class BaseRouter:
    def __init__(self):
        self.registry = []

    def register(
        self,
        viewset,
        prefix=None,
        basename=None,
        trailing_slash=True,
    ):
        if basename is None:
            basename = self.get_default_basename(viewset)
        if prefix is None:
            prefix = basename
        if trailing_slash and not prefix.endswith("/"):
            prefix = "%s/" % prefix
        self.registry.append((prefix, viewset, basename))

        # invalidate the urls cache
        if hasattr(self, "_urls"):
            del self._urls

    def get_default_basename(self, viewset):
        """
        If `basename` is not specified, attempt to automatically determine
        it from the viewset.
        """
        raise NotImplementedError("get_default_basename must be overridden")

    def get_urls(self):
        """
        Return a list of URL patterns, given the registered viewsets.
        """
        raise NotImplementedError("get_urls must be overridden")

    @property
    def urls(self):
        if not hasattr(self, "_urls"):
            self._urls = self.get_urls()
        return self._urls


class SimpleRouter(BaseRouter):
    def get_urls(self):
        """
        Use the registered viewsets to generate a list of URL patterns.
        """
        urls = []
        for prefix, viewset, name in self.registry:
            urls.append(path(prefix, include(viewset.urls)))
        return urls

    def get_default_basename(self, viewset):
        """
        If `basename` is not specified, attempt to automatically determine
        it from the viewset.
        """
        model = getattr(viewset, "model", None)

        assert model is not None, (
            "`basename` argument not specified, and could "
            "not automatically determine the name from the viewset, as "
            "it does not have a `.model` attribute."
        )

        return model._meta.object_name.lower()


class DefaultIndexView(SiteView):
    template_name = "index.html"

    def get_template_names(self):
        return super().get_template_names()


class DefaultRouter(SimpleRouter):
    """
    The default router extends the SimpleRouter, but also adds in a default
    API root view, and adds format suffix patterns to the URLs.
    """

    index_enabled = True
    index_view_name = "index"
    index_view_class = DefaultIndexView

    def get_index_view_class(self):
        """
        Return a basic root view.
        """
        return self.index_view_class

    def get_urls(self):
        """
        Generate the list of URL patterns, including a default root view
        for the API, and appending `.json` style format suffixes.
        """
        urls = super().get_urls()
        if self.index_enabled:
            urls += (
                path(
                    "",
                    self.get_index_view_class().as_view(),
                    name=self.index_view_name,
                ),
            )
        return urls


class AuthenticationRouter(SimpleRouter):
    def get_urls(self):
        try:
            from allauth.urls import urlpatterns

            urls = super().get_urls()
            urls += list(urlpatterns)
            return urls
        except Exception as err:
            return []


class Site(DefaultRouter):

    site_url = "/"
    site_view_hook_name = "REGISTER_SITE_VIEW"
    site_path_hook_name = "REGISTER_SITE_PATH"
    authentication_router_class = AuthenticationRouter

    def __init__(self):
        super().__init__()

    def each_context(self, request):
        """
        Return a dictionary of variables to put in the template context for
        *every* page in the admin site.

        For sites running on a subpath, use the SCRIPT_NAME value if site_url
        hasn't been customized.
        """
        script_name = request.META["SCRIPT_NAME"]
        site_url = script_name if routers_settings.SITE_URL == "/" and script_name else self.site_url
        return {
            "site_title": routers_settings.SITE_TITLE,
            "site_header": routers_settings.SITE_HEADER,
            "site_url": site_url,
        }

    def index_view(self, request):
        context = self.each_context(request)
        context.update(
            {
                "title": routers_settings.INDEX_TITLE,
            }
        )
        return self.get_index_view_class().as_view(extra_context=context)(request)

    def get_authentication_router(self):
        return self.authentication_router_class()

    def get_authentication_urls(self):
        return self.get_authentication_router().get_urls()

    def get_hooked_views(self):
        # Get registered custom admin view
        funcs = hookup.get_hooks(self.site_view_hook_name)
        urls = []
        for func in funcs:
            url_path, view, name = func()
            if isclass(view):
                if not issubclass(view, View):
                    raise ImproperlyConfigured("%s must be subclass of View" % view)
                route = path(url_path, view.as_view(), name=name)
            elif callable(view):
                route = path(url_path, view, name=name)
            else:
                raise ImproperlyConfigured("%s must be View or function" % view)
            urls.append(route)
        return urls

    def get_hooked_paths(self):
        urls = []
        funcs = hookup.get_hooks(self.site_path_hook_name)
        for func in funcs:
            urls.append(func())
        return urls

    def get_urls(self):
        """
        Generate the list of URL patterns, including a default root view
        for the API, and appending `.json` style format suffixes.
        """
        urls = super().get_urls()
        urls += self.get_authentication_urls()
        urls += self.get_hooked_views()
        urls += self.get_hooked_paths()

        return urls



from django.apps import AppConfig


class SimpelSitesConfig(AppConfig):
    icon = "web"
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_routes"
    label = "django_routes"
    verbose_name = "Django Routes"



from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model

from .helpers import ButtonHelper, PermissionHelper, URLHelper


class GenericViewSet:

    icon = ""
    model = None
    menu_label = None
    menu_icon = None
    menu_order = None
    # Helper class

    url_helper_class = URLHelper
    button_helper_class = ButtonHelper
    permission_helper_class = PermissionHelper
    template_namespace = None

    def __init__(self, parent=None):
        """Don't allow initialisation unless self.model is set to a valid model"""
        if not self.model or not issubclass(self.model, Model):
            raise ImproperlyConfigured(
                "The model attribute on your '%s' class must be set, and "
                "must be a valid Django model." % self.__class__.__name__
            )
        self.parent = parent
        self.opts = self.model._meta
        self.url_helper = self.get_url_helper_class()(self.model)
        self.permission_helper = self.get_permission_helper_class()(self.model)

    def get_queryset(self, request):
        """Returns a QuerySet of all model instances."""
        qs = self.model._default_manager.get_queryset()
        return qs

    def get_permission_helper_class(self):
        """Returns a permission_helper class to help with permission-based logic."""
        return self.permission_helper_class

    def get_permission_helper(self):
        return self.permission_helper_class(self.model)

    def get_button_helper_class(self):
        """Returns a ButtonHelper class to help generate buttons for the given model."""
        return self.button_helper_class

    def get_button_helper(self, view, request):
        return self.button_helper_class(view, request)

    def get_url_helper_class(self):
        return self.url_helper_class

    def get_url_helper(self):
        return self.url_helper_class(self.model)

    def get_menu_label(self):
        """Returns the label text to be used for the menu item."""
        return self.menu_label or self.opts.verbose_name_plural.title()

    def get_menu_icon(self):
        return self.menu_icon

    def get_menu_order(self):
        """
        Returns the 'order' to be applied to the menu item. 000 being first
        place. Where ViewSetGroup is used, the menu_order value should be
        applied to that, and any ModelAdmin classes added to 'items'
        attribute will be ordered automatically, based on their order in that
        sequence.
        """
        return self.menu_order or 999

    def get_templates(self, action="index"):
        """
        Utility function that provides a list of templates to try for a given
        view, when the template isn't overridden by one of the template
        attributes on the class.
        """
        namespace = self.get_template_namespace()
        app_label = self.opts.app_label.lower()
        model_name = self.opts.model_name.lower()
        if namespace:
            return [
                "%s/%s/%s/%s.html" % (namespace, app_label, model_name, action),
                "%s/%s/%s.html" % (namespace, app_label, action),
                "%s/%s.html" % (namespace, action),
            ]
        else:
            return [
                "%s/%s/%s.html" % (app_label, model_name, action),
                "%s/%s.html" % (app_label, action),
                "%s.html" % (action),
            ]

    def get_template_namespace(self):
        return self.template_namespace

    def get_urls(self):
        """Utilised by 'register_webapp_urls' hook to register urls to router."""
        return tuple()

    def has_view(self, view_name):
        return hasattr(self, view_name)
